#include <stdlib.h>
#include <string>

//#define __USE_GNU 

#include <math.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <list>
#include <algorithm>
#include <gsl/gsl_sf_dilog.h>

#include "defs.h"
#include "templates.cpp"
#include "geodesic_boundary.h"
#include "list.cpp"
#include "surfaces.h"


bool scrivi_dati_solo_se_soluzione_geometrica = TRUE;
bool dati_theta_dettagliati = FALSE;
long double tolleranza_F = 0.0000001;
long double tolleranza_tilt = 0.000001;
long double tolleranza_theta = 0.00000000000001;
double tolleranza_volume = 0.000000001;
long double raggio_sezione = 0.00001;
int scrivi_angoli_cifre_significative = 10;

//considerare gli angoli tra gli invarianti ?
bool con_angoli = FALSE;


int inout_to_theta[4][4] = {{-1,1,0,2},{1,-1,5,3},{0,5,-1,4},{2,3,4,-1}};
int bordo_theta[4][3] = {{3,1,2}, {5,0,1}, {3,4,5}, {2,0,4}};

extern bool inglese;

//BORDO:                       
//
//       3                     0
//       |                   5   1
//    1  |  5              
//       |                2         4
//  0----------         3             3
//       |                1         5
//    2  |  4      
//       |                   4   2
//                             0       
// inout_to_theta          bordo_theta
//
//

//costruttore standard
tvertice::tvertice ():overtice() {
  for (int t=0; t<6; t++) theta[t]=-1;
  for (int t=0; t<4; t++) circo_raggio[t]=-1;
  for (int t=0; t<4; t++) cusp[t]=FALSE;
}

//tvertice = overtice con theta = -1
tvertice::tvertice (const overtice& ov):overtice(ov) {
  for (int t=0; t<6; t++) theta[t]=-1;
  for (int t=0; t<4; t++) circo_raggio[t]=-1;
  for (int t=0; t<4; t++) cusp[t]=FALSE;
}

//tgrafo = ografo con theta = -1
tgrafo::tgrafo (const ografo& o) {
  const int l = o.lung();
  for (int i=0; i<l; i++)
    v.push_back(new tvertice(*((overtice*) o[i])));
  ff = o.get_ff();
  set_bordo();
}

//costruttore di copia
tgrafo::tgrafo (const tgrafo& o):bd(o.bd), s(o.s), ha_cuspidi(o.ha_cuspidi) {
  const int l = o.lung();
  for (int i=0; i<l; i++)
    v.push_back(new tvertice(*((tvertice*) o[i])));
  ff = o.get_ff();
}

//carica tgrafo dal file
tgrafo::tgrafo (std::istream& file, int tipo_formato):bd(), s() {
  *this = tgrafo(ografo(file, tipo_formato));
  set_bordo();
  if (tipo_formato >= 1) return;
  const int l = lung();
  char w[1000];
  file.seekg(-2, std::ios::cur);
  file.read(w,1);
  //se sono scritti anche gli angoli, li carica:
  //ATTENZIONE: scrivere routine con precisione arbitraria in ingresso
  if (w[0]!='-') {
    file.getline(w,800);
    for (int t=0; t<l; t++) {
      file.getline(w,800);
      for (int t2=0; t2<6; t2++)
        set_theta(t,t2,-1); //atof(&w[9+15*t2]));  //atod non esiste!!
    }
  }
  file.getline(w,100);   //stringa '--------' finale
}

//assegnamento
tgrafo& tgrafo::operator=(const tgrafo& o) {
  disalloca(v);
  v = std::vector<vertice*>(0);
  for (int i=0; i<o.lung(); i++)
    v.push_back(new tvertice(*((tvertice*) o[i])));
  ff = o.get_ff();
  bd = o.bd;
  s = o.s;
  ha_cuspidi = o.ha_cuspidi;
  return *this;
}

void tgrafo::scrivi_theta(std::ostream& f) const {
  f << (inglese ? "Angles :\n" : "Angoli :\n");
  for (int i=0; i<lung(); i++) {
    f << "N. " << std::setw(2) << i << ":   ";
    for (int j=0; j<6; j++) {
      f.setf(std::ios::fixed); 
      f.precision(scrivi_angoli_cifre_significative);
      f << get_theta(i,j);
      f << "   ";
    }
    f << "\n";
  }
}

void tvertice::scrivi_circo_raggio(std::ostream& f) const {
  for (int i=0; i<4; i++) 
    f << circo_raggio[i] << "   ";
  f << std::endl;
}

void tgrafo::scrivi_circo_raggio(std::ostream& f) const {
  std::cout << "Circo raggi: " << std::endl;
  for (int i=0; i<lung(); i++) {
    f << "N. " << i << ":  ";
    ((tvertice* )v[i])->scrivi_circo_raggio(f);
  }
}

//scrive gli invarianti geometrici del tgrafo
void tgrafo::scrivi_geometria(std::ostream& f) const {
  if (scrivi_dati_solo_se_soluzione_geometrica && !soluzione_geometrica()) return;
  f << "Volume: " << volume() << "  (";
  for (int t=0; t<lung(); t++) {
    f << ((tvertice*) v[t])->volume();
    (t<lung()-1) ? f << ", " : f << ")\n";
  }
  if (canonica()) f << (inglese ? "The triangulation is canonical.\n" : 
			"La triangolazione e' canonica.\n");
  else {
    if (sub_canonica()) f << (inglese ? 
			      "The canonical decomposition consists of a regular octahedron, and is obtained from this triangulation by merging together some tetrahedra.\n" : "La triangolazione e' subcanonica.\n");
    else {
      f << "Tilt: ";
      std::vector<double> tl = tilt();
      for (int t=0; t<tl.size(); t++) {
	f << tl[t] << "  ";
	if (!(t % 6)) f << std::endl;
      }
      f << "\n";
    }
  }
  if (dati_theta_dettagliati) {
    //scrivi_lunghezze_spigoli(f);  ATTENZIONE cambiato qui
    scriviF(f, ha_cuspidi ? get_F_boundary() : get_F());
    //scrivi_circo_raggio(f);  ATTENZIONE cambiato qui
  }
}

//oltre alle facce, calcola il bordo e vede se ci sono cuspidi
void tgrafo::set_facce() {
  ografo::set_facce();
  set_bordo();
}

//calcola il bordo e vede se ci sono cuspidi
void tgrafo::set_bordo() {
  bd = grafo3_di_bordo(*this);
  s = superficie(bd);
  ha_cuspidi = FALSE;
  for (int i=0; i<s.componenti_connesse(); i++)
    if (s.get_eulero(i) == 0)
      ha_cuspidi =  TRUE;
  const int l=lung();
  if (!ha_cuspidi) for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      set_cusp(i,j,0);
  else {
    std::vector<int> cuspidi = bd.vertici_in_componenti_connesse();
    for (int i=0; i<l; i++)
      for (int j=0; j<4; j++)
        set_cusp(i,j,(s.get_eulero(cuspidi[4*i+j]-1) == 0));
  }
}

//determina una oro-sezione per ogni componente di bordo i-esima (partendo da 0)
//con ampiezza determinata da r[i]
void tgrafo::espandi_raggi(const std::vector<double>& r) const {
  const int l = lung();
  std::vector<int> cc = bd.vertici_in_componenti_connesse();
  int num_toro = 0;
  for (int c=0; c<s.componenti_connesse(); c++)
    if (s.get_eulero(c) == 0) {
      int vt=0;
      while(cc[vt] != c+1) vt++;
      set_circo_raggio(vt/4, vt%4, r[num_toro++]);
      bool ho_finito = FALSE;
      while (!ho_finito) {
        ho_finito = TRUE;
        for (int i=0; i<4*l; i++)
          if (cc[i] == c+1 && get_circo_raggio(i/4, i%4) < 0) {
            ho_finito = FALSE;
            int j = 0;
            for (; j<3 && get_circo_raggio(bd.get_v(i,j) / 4,bd.get_v(i,j) % 4)<0; j++) {}
            if (j!=3) {
              int i2 = bd.get_v(i, j);
              int j2 = bd.get_col(i, j);
              int v2 = i2 / 4;
              int c2 = i2 % 4;
              int v = i / 4;
              int c = i % 4;
              double r = get_circo_raggio(v2,c2) * sin(get_theta(v2, bordo_theta[c2][j2])) /
                sin(get_theta(v, bordo_theta[c][j]));
              set_circo_raggio(v,c,r);
            }
          }
      }
    }
}

//scrive il tgrafo su f
//scrive gli angoli solo se la soluzione e' geometrica 
//(con tollerana_soluzione_geometrica)
//dipende dalla variabile globale 'dati_theta_dettagliati'
//se dati_theta_dettagliati == TRUE,
//   scrive gli angoli theta e varie grandezze geometriche
void tgrafo::scrivi(std::ostream& f, int tipo_formato) const {
  ografo::scrivi(f, tipo_formato);
  if (scrivi_dati_solo_se_soluzione_geometrica && !soluzione_geometrica()) return;
  scrivi_theta(f);
}

//salva su file con formato 'tipo_formato' (vedi ograph.cpp)
void tgrafo::salva(std::ostream& f, int tipo_formato) const {
  int prec_dati_theta_dettagliati = dati_theta_dettagliati;
  dati_theta_dettagliati = FALSE;
  scrivi(f, tipo_formato);
  f << "------------\n";
  dati_theta_dettagliati = prec_dati_theta_dettagliati;
}

//mette tutti i theta uguali a -1
void tgrafo::reset_theta() {
  for (int i=0; v[i]; i++)
    for (int j=0; j<6; j++)
      set_theta(i,j,-1);
}

long double tvertice::cos(int i, int j) const {
  if (i==j) std::cerr << "Errore!";
  long double ris = ::cosl(theta[inout_to_theta[i][j]]);
  return ris;
}

long double tvertice::sin(int i, int j) const {
  if (i==j) std::cerr << "Errore!";
  long double ris = ::sinl(theta[inout_to_theta[i][j]]);
  return ris;
}

long double tvertice::c(int i, int j, int a, int b) const {
  long double ris = cos(i,j) * (cos(i,a) * cos(j,b) + cos(i,b) * cos(j,a)) +
    cos(i,a) * cos(i,b) + cos(j,a) * cos(j,b) +
    cos(a,b) * sin(i,j) * sin(i,j);
  return ris;
}

//calcola la derivata di c(i,j,a,b) tispetto a theta[inout_to_theta(r,s)]
long double tvertice::dc(int i, int j, int a, int b, int r, int s) const {
  int quanti_coincidono = 0;
  if (r==i || r==j) quanti_coincidono++;
  if (s==i || s==j) quanti_coincidono++;
  long double ris = 0;
  if (quanti_coincidono == 2)
    ris = - sin(i,j) * (cos(i,a) * cos(j,b) + cos(i,b) * cos(j,a))
      + 2 * cos(a,b) * cos(i,j) * sin(i,j);
  if (quanti_coincidono == 1) {  
    //mette in modo che r=i, s=a:
    if (s==i || s==j) std::swap(r, s);
    if (r==j) std::swap(i, j);
    if (s==b) std::swap(a, b);
    ris = - sin(i,a) * cos(i,j) * cos(j,b) - sin(i,a) * cos(i,b);
  }
  if (quanti_coincidono == 0)
    ris = - sin(a,b) * powl(sin(i,j),2);
  return ris;
}

long double tvertice::d(int i, int j, int a) const {
  long double ris1 = 2 * cos(i,j) * cos(j,a) * cos(a,i);
  long double ris2 = powl(cos(i,j),2);
  long double ris3 = powl(cos(j,a),2);
  long double ris4 = powl(cos(a,i),2);
  return (ris1+ris2+ris3+ris4-1);
}

//calcola la derivata di d(i,j,a) tispetto a theta[inout_to_theta(r,s)]
long double tvertice::dd(int i, int j, int a, int r, int s) const {
  long double ris = 0;
  //se {r,s} non e' contenuto in {i,j,a} allora ris=0
  if ((r==i || r==j || r==a) && (s==i || s==j || s==a)) {
    //25/09/02: modificato da qui in poi
    long double add = 0;
    if (r==i && s==j || r==j && s==i) add = cos(j,a) * cos(a,i);
    if (r==j && s==a || r==a && s==j) add = cos(a,i) * cos(i,j);
    if (r==a && s==i || r==i && s==a) add = cos(i,j) * cos(j,a);
    ris = - 2 * sin(r,s) * add - 2 * sin(r,s) * cos(r,s);
  }
  return ris;
}

//calcola l rispetto a theta[inout_to_theta(i,j)]
long double tvertice::l(int i, int j) const {
  int a=-1; int b=-1;
  for (int t=0; t<4; t++)
    if (t!=i && t!=j) 
      a>=0 ? b=t : a=t;
  long double numeratore = powl(c(i,j,a,b),2);
  long double denominatore = d(i,j,a) * d(i,j,b);
  return (numeratore/denominatore);
}

//calcola la derivata di l(i,j) rispetto a theta[inout_to_theta(r,s)]
//(le derivate rispetto ai theta sugli altri punti sono zero)
long double tvertice::dl(int i, int j, int r, int s) const {
  int a=-1; int b=-1;
  for (int t=0; t<4; t++)
    if (t!=i && t!=j) 
      a>=0 ? b=t : a=t;
  long double numeratore = 2 * c(i,j,a,b) * dc(i,j,a,b,r,s);
  long double denominatore = d(i,j,a) * d(i,j,b);
  long double num_addendo1 = dd(i,j,a,r,s);
  long double den_addendo1 = powl(d(i,j,a),2) * d(i,j,b);
  long double num_addendo2 = dd(i,j,b,r,s);
  long double den_addendo2 = powl(d(i,j,b),2) * d(i,j,a);
  long double risposta = numeratore/denominatore - powl(c(i,j,a,b),2) *
    (num_addendo1/den_addendo1 + num_addendo2/den_addendo2);
  return risposta;
}

//restituisce F[0], ... ,F[6*(l+1)-1], che sono i valori che
//debbono annullarsi se le theta rappresentano 
//una triangolazione troncata iperbolica
//In particolare, se f e' la prima faccia: F[0] = somma_theta(0)-2*PI
//e per 1 <= i <f->lati()-1 abbiamo 
//F[i] = passaggio[i-1]->l(entrata[i-1],uscita[i-1]) - passaggio[i]->l(entrata[i],uscita[i]),
//e cosi' via per le altre facce.
std::vector<long double> tgrafo::get_F() const {
  std::vector<long double> F;
  int indice=0;
  for (int i = 0; i<ff.size(); i++) {
    F.push_back(somma_theta(i) - 2*M_PIl);
    for (int t=0; t<ff[i].lati()-1; t++)
      F.push_back(get_l(i,t+1) - get_l(i,t));
  }
  return F;
}

//restituisce F[0], ..., F[quante_facce - 1], con F[i] = somma_theta(i) - 2*PI
//F[quante_facce], ... ,F[quante_facce + 6*lung()], 
//con F[i] = differenza fra lunghezze spigolo di bordo visto dai due estremi
std::vector<long double> tgrafo::get_F_boundary() const {
  const int l = lung();
  std::vector<long double> F;
  for (int i = 0; i<ff.size(); i++)
    F.push_back(somma_theta(i) - 2*M_PIl);
  for (int i = 0; i<4*l; i++)
    for (int j = 0; j<3; j++)
      if (bd.prima_occorrenza_spigolo(i,j)) {
	int i2 = bd.get_v(i,j);
	int j2 = bd.get_col(i,j);
	F.push_back(get_lb(i,j) - get_lb(i2,j2));
      }
  return F;
}

//restituisce una matrice 6(l+1) per 6(l+1) che rappresenta il
//Jacobiano di F rispetto alle theta, ordinate per v
//(cioe' la 6*i+j -esima e' la j-esima di v[i])
std::vector<std::vector<long double> > tgrafo::dF() const {
  int righe = 6*lung();
  std::vector<std::vector<long double> > J(righe, std::vector<long double> (righe,0));
  int indice=0;
  for (int i = 0; i<ff.size(); i++) {
    const faccia& f=ff[i];
    for (int t=0; t<f.lati(); t++) {
      int num_theta = 6*(f.get_passaggio(t)) + 
	inout_to_theta[f.get_entrata(t)][f.get_uscita(t)];
      J[indice][num_theta] = 1;
    }
    indice++;
    for (int t=0; t<f.lati()-1; t++) {
      for (int r=0; r<4; r++)
        for (int s=r+1; s<4; s++) {
          int num_theta = 6*(f.get_passaggio(t)) + inout_to_theta[r][s];
          long double dl = 
	    get_dl(f.get_passaggio(t), f.get_entrata(t), f.get_uscita(t),r,s);
          J[indice][num_theta] -= dl;  //attenzione: segno non sicuro
        }
      for (int r=0; r<4; r++)
        for (int s=r+1; s<4; s++) {
          int num_theta = 6*(f.get_passaggio(t+1)) + inout_to_theta[r][s];
          long double dl = get_dl(f.get_passaggio(t+1), 
				  f.get_entrata(t+1), f.get_uscita(t+1),r,s);
          J[indice][num_theta] += dl;  //attenzione: segno non sicuro
        }
      indice++;
    }
  }
  return J;
}

void tgrafo::scriviF(std::ostream& f, const std::vector<long double>& F) const {
  f << "Funzione F:\n";
  for (int t=0; t<F.size(); t++) {
    f << F[t] << "  ";
    if (t && !(t % 8)) f << "\n";
  }
  f << "\n";
}

void tgrafo::scriviX(std::ostream& f, const std::vector<long double>& X) const {
  int righe = 6*lung();
  f << "Soluzione X:\n";
  for (int t=0; t<righe; t++) f << X[t] << "  ";
  f << "\n";
}

void tgrafo::scriviJ(std::ostream& f, const std::vector<std::vector<long double> >& J) const {
  int righe = 6*lung();
  f << "\nJacobiano:\n";
  for (int t=0; t<righe; t++) {
    for (int t2=0; t2<righe; t2++)
      f << std::setw(8) << J[t][t2] << "  ";
    f << "\n";
  }
}

//restituisce il determinante di una matrice
//funziona solo se e' quadrata
long double determinant(std::vector<std::vector<long double> > A) {
  int sz = A.size();
  if (sz == 1) return A[0][0];
  long double res = 0;
  for (int i=0; i<sz; i++) {
    std::vector<std::vector<long double> > B;
    for (int j=1; j<sz; j++) {
      std::vector<long double> riga;
      for (int k=0; k<sz; k++) 
	if (k!=i) riga.push_back(A[j][k]);
      B.push_back(riga);
    }
    res+=(long double) powl(-1,i)*A[0][i]*determinant(B);
  }
  return res;
}

//restituisce la matrice di Gram
std::vector<std::vector<long double> > tvertice::gram() const {
  std::vector<std::vector<long double> > A(4,std::vector<long double>(4,0));
  for (int i=0; i<4; i++) 
    for (int j=0; j<4; j++) {
      if (i==j) A[i][j] = 1;
      else A[i][j] = -::cosl(theta[inout_to_theta[i][j]]);
    }
  return A;
}

//restituisce funzione U che serve per il volume
std::complex<double> tvertice::U(std::complex<double> z) const {
  std::complex<double> ris = dilog(z);
  std::complex<double> I(0,1);
  for (int t=0; t<3; t++) {
    std::complex<double> argomento(1);
    for (int s=0; s<6; s++) if (s!=t && s!=t+3)
      argomento *= (::cos(theta[s]) + I*::sin(theta[s]));
    argomento *= z;
    std::complex<double> dilogarg = dilog(argomento);
    ris += dilogarg;
  }
  for (int i=0; i<4; i++) {
    std::complex<double> argomento(-1);
    for (int j=0; j<4; j++) if (i!=j) {
      int t = (inout_to_theta[i][j] + 3) % 6;
      argomento *= (::cos(theta[t]) + I*::sin(theta[t]));
    }
    argomento *= z;
    ris -= dilog(argomento);
  }
  ris /= 2;
  return ris;
}

//restituisce il volume del tetraedro troncato con gli angoli theta
double tvertice::volume() const {
  std::vector<std::vector<long double> > A = gram();
  double det = determinant(A);
  std::complex<double> num1 = 0;
  std::complex<double> I(0,1);
  for (int t=0; t<3; t++)
    num1 += ::sin(theta[t]) * ::sin(theta[t+3]);
  double det_radice = sqrt(-det);
  std::complex<double> den = 0;
  for (int t=0; t<3; t++)
    den += (::cos(theta[t]) + I*::sin(theta[t])) *
      (::cos(theta[t+3]) + I*::sin(theta[t+3]));
  for (int i=0; i<4; i++) {
    std::complex<double> addendo = 1;
    for (int j=0; j<4; j++) if (i!=j) {
      int t = inout_to_theta[i][j];
      addendo *= (::cos(theta[t]) + I*::sin(theta[t]));
    }
    den += addendo;
  }
  std::complex<double> addendo = 1;
  for (int t=0; t<6; t++)
    addendo *= (::cos(theta[t]) + I*::sin(theta[t]));
  den += addendo;
  std::complex<double> due = 2;
  std::complex<double> z1 = -due * (num1 - det_radice * I) / den;
  std::complex<double> z2 = -due * (num1 + det_radice * I) / den;
  std::complex<double> Udiff = U(z1) - U(z2);
  double ris = Udiff.imag();
  ris /= 2;
  return ris;
}

//restituisce il volume della varieta'
double tgrafo::volume() const {
  double tot = 0;
  const int l = lung();
  for (int i = 0; i < l; i++) 
    tot+=((tvertice*) v[i])->volume();
  return tot;
}

//dice se le theta sono una soluzione, 
//a meno di 'tolleranza_F' (ove 'tolleranza_F'>0)
//ovvero dice se |F[i]|<tolleranza_F per ogni i
bool tgrafo::soluzione_geometrica() const {
  if (!soluzione_nel_dominio()) return FALSE;
  std::vector<long double> F = (ha_cuspidi ? get_F_boundary() : get_F());
  for (int i=0; i<F.size(); i++) {
    if (!finite(F[i])) return FALSE;
    if (abs(F[i])>tolleranza_F) return FALSE;
  }
  return TRUE;
}

//dice se la soluzione e' nel dominio
bool tgrafo::soluzione_nel_dominio(int num_vertice) const {
  if (num_vertice >= 0) 
    return (((tvertice*) v[num_vertice])->soluzione_nel_dominio());
  const int l=lung();
  for (int i=0; i<l; i++)
    if (!((tvertice*) v[i])->soluzione_nel_dominio()) return FALSE;
  return TRUE;
}

//dice se la soluzione del singolo vertice e' nel dominio
bool tvertice::soluzione_nel_dominio() const {
  for (int j=0; j<6; j++)
    if (get_theta(j)<0) return FALSE;
  for (int j=0; j<4; j++) {
    long double tot = 0;
    for (int k=0; k<4; k++)
      if (j!=k) tot+=get_theta((inout_to_theta[j][k] + 3) % 6);
    if (tot>=M_PIl) return FALSE;
  }
  return TRUE;
}

//funzione g()
double tvertice::g() const {
  double ris = -1;
  for (int t = 0; t<6 ; t++)
    ris += pow(::cos(theta[t]) , 2);
  for (int i = 0; i<4 ; i++) {
    double addendo = 2;
    for (int j = 0; j<4 ; j++) if (i!=j) {
      int t = (inout_to_theta[i][j] + 3) % 6;
      addendo *= ::cos(theta[t]);
    }
    ris += addendo;
  }
  for (int t = 0; t<3; t++) {
    double addendo = 2;
    for (int s = 0; s<6; s++) if (s!=t && s!=t+3)
      addendo *= ::cos(theta[s]);
    ris += addendo;
  }
  for (int t = 0; t<3; t++)
    ris -= pow(::cos(theta[t]) * ::cos(theta[t+3]),2);
  return ris;
}

//funzione D()
double tvertice::D(int ii) const {
  int ie = (ii+2) % 4; //da intrinseco a estrinseco
  if (cusp[ie]) {
    double num = 0;
    double den = 1;
    for (int j = 1; j<=3; j++) {
      int i2 = (ii+j) % 4;
      int t = inout_to_theta[ii][i2];
      int t2 = (t+3) % 6;
      num += ::cos(theta[t]) * ::sin(theta[t2]);
      den *= ::sin(theta[t2]);
    }
    return (num / (2 * den * circo_raggio[ie]));
  }
  else return sqrt(g()/d((ii+1) % 4, (ii+2) % 4, (ii+3) % 4));
}
	
double tvertice::tilt(int i) const {
  std::vector<std::vector<long double> > G = gram();
  double ris = 0;
  for (int j=0; j<4; j++) 
    ris += G[i][j] / D(j);
  return ris;
}

std::vector<double> tgrafo::tilt() const {
  std::vector<double> ris;
  const int l = lung();
  if (ha_cuspidi) {
    std::vector<double> r;
    for (int i=0; i<s.componenti_connesse(); i++)
      if (s.get_eulero(i) == 0)
        r.push_back(raggio_sezione);
    espandi_raggi(r);
  }
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      if (prima_occorrenza_spigolo(i,j)) {
	int i2 = get_v(i,j);
	int j2 = get_col(i,j);
	double tl = ((tvertice* ) 
		     v[i])->tilt(j) + ((tvertice* ) v[i2])->tilt(j2);
	ris.push_back(tl);
      }
  return ris;
}

bool tgrafo::canonica() const {
  if (!soluzione_geometrica()) return FALSE;
  std::vector<double> tl = tilt();
  const int l = lung();
  for (int t=0; t<2*l; t++)
    if (tl[t]>=-tolleranza_tilt) return FALSE;
  return TRUE;
}

bool tgrafo::sub_canonica() const {
  if (!soluzione_geometrica()) return FALSE;
  std::vector<double> tl = tilt();
  const int l = lung();
  for (int t=0; t<2*l; t++)
    if (tl[t]>=tolleranza_tilt) return FALSE;
  return TRUE;
}

void tgrafo::newton(int quante_volte, std::ostream& f,
		    bool scrivi_dati) {
  for (int n=0; n<ff.size(); n++) {
    const faccia& f = ff[n];
    for (int i=0; i<f.lati(); i++)
      set_theta(f.get_passaggio(i), 
		inout_to_theta[f.get_entrata(i)][f.get_uscita(i)], 
		//M_PIl*griglia /(5*max_griglia));
		//M_PIl/f.lati());
		0.5 + 0.01*i);
  }
  if (scrivi_dati) scrivi_theta(f);
  for (int t=0; t<quante_volte && !soluzione_geometrica(); t++) 
    if (!newton(scrivi_dati, f)) break;
  if (scrivi_dati) f << "--------------\n";
  if (soluzione_geometrica()) return;
}

bool tgrafo::newton(bool scrivi_dati, std::ostream& f) {
  const int l=lung();
  int righe = 6*l;
  std::vector<long double> F = get_F();
  std::vector<std::vector<long double> > J = dF();
  std::vector<long double> X;
  
  if (scrivi_dati) scriviF(std::cout, F); 
  
  //l'equazione e' J*X = F
  //trasforma in triangolare superiore
  for (int j=0; j<righe; j++) {
    int pivot=j;
    for (int i=j; i<righe; i++) if (abs(J[i][j])>abs(J[pivot][j])) pivot=i;
    for (int j2=0; j2<righe; j2++) std::swap(J[pivot][j2],J[j][j2]);
    std::swap(F[pivot],F[j]);
    for (int i=j+1; i<righe; i++) {
      long double k=-J[i][j]/J[j][j];
      for (int j2=j; j2<righe; j2++)
        J[i][j2]+=(k*J[j][j2]);
      F[i]+=k*F[j];
    }
  }

  //trasforma in diagonale
  for (int j=righe-1; j>=0; j--)
    for (int i=j-1; i>=0; i--) {
      F[i]-=(J[i][j]/J[j][j])*F[j];
    }

  //evita 'nan' e 'inf'
  for (int i=0; i<righe; i++)
    if (!finite(F[i])) return FALSE;
  
  //scrive soluzioni
  for (int i=0; i<righe; i++)
    X.push_back(F[i]/J[i][i]);

  //theta' (il nuovo theta) sara' theta-scalamento*X
  long double scalamento = 1;

  //salva i theta attuali
  std::vector<std::vector<long double> > th_old(l, std::vector<long double>(6,0));
  for (int r=0; r<l; r++)
    for (int s=0; s<6; s++)
      th_old[r][s] = get_theta(r,s);
  
  do {
    
    //cambia theta
    for (int r=0; r<l; r++)
      for (int s=0; s<6; s++)
	set_theta(r,s,th_old[r][s] - scalamento*X[6*r+s]);
    
    //controlla che il nuovo theta sia nel dominio, e 
    //piu' vicino al vecchio theta che al bordo del dominio.
    //Se non e' cosi', sceglie uno 'scalamento' in modo che sia cosi'
    scalamento/=2;
    
  } while (!soluzione_nel_dominio());
  
  
  //se vede che non c'e' piu' cambiamento, esce
  bool sono_cambiate = FALSE;
  for (int r=0; r<l && !sono_cambiate; r++)
    for (int s=0; s<6; s++)
      if (abs(th_old[r][s] - get_theta(r,s))>tolleranza_theta) sono_cambiate = TRUE;
  if (!sono_cambiate)
    return FALSE;
  
  if (scrivi_dati) {
    scrivi_theta(f);
    scrivi_lunghezze_spigoli(f);
  }
  return TRUE;
}

//calcola la somma dei theta sugli angoli della faccia
long double tgrafo::somma_theta(int num_faccia) const {
  long double somma=0;
  const faccia& f = ff[num_faccia];
  for (int i=0; i<f.lati(); i++)
    somma+=get_theta(f.get_passaggio(i), 
		     inout_to_theta[f.get_entrata(i)][f.get_uscita(i)]);
  return somma;
}

//restituisce la derivata di l(i,j) tispetto a theta[inout_to_theta(r,s)]
//nel vertice 'vert'
long double tgrafo::get_dl(int vert, int i, int j, int r, int s) const {
  return ((tvertice*) v[vert])->dl(i,j,r,s);
}

//restituisce la lunghezza del lato (della triangolazione duale) 
//associato alla faccia, vista dal vertice t-esimo (0<=t<lati)
long double tgrafo::get_l(int num_faccia, int t) const {
  const faccia& f = ff[num_faccia];
  const tvertice* vert = (tvertice*) v[f.get_passaggio(t)];
  return (vert->l(f.get_entrata(t), f.get_uscita(t)));
}

//restituisce la lunghezza del lato della triangolazione
//nel bordo, associato al vertice3 'num_vert3', uscita 'usc3'
long double tgrafo::get_lb(int num_vert3, int usc3) const {
  int vert = num_vert3 / 4;
  int usc = num_vert3 % 4;
	
  double th[3];
  for (int i=0; i<3; i++)
    th[i] = get_theta(vert, bordo_theta[usc][(usc3+i) % 3]);
	
  long double ris = (cosl(th[1]) * cosl(th[2]) + cosl(th[0])) / 
    (sinl(th[1]) * sinl(th[2]));
  return ris;
}

//scrive le lunghezze degli spigoli (duali alle facce della spina)
void tgrafo::scrivi_lunghezze_spigoli(std::ostream& fstr) const {
  fstr << "Lunghezze degli spigoli (duali):\n";
  for (int num_faccia = 0; num_faccia<ff.size(); num_faccia++) {
    fstr << "N. " << num_faccia+1 << ": ";
    for (int t=0; t<ff[num_faccia].lati(); t++) {
      if (t && !(t % 6)) fstr << "\n";
      fstr << std::setw(8) << get_l(num_faccia, t) << "  ";
    }
    fstr << "\n";
  }
}

//restituisce il dilogaritmo
std::complex<double> dilog(const std::complex<double>& z) {
  gsl_sf_result re; gsl_sf_result im;
  gsl_sf_complex_dilog_e(norm(z), arg(z), &re, &im);
  std::complex<double> ris(re.val ,im.val);
  return ris;
}

//cerca sub_canonica, se la trova trasforma l'ografo in questa
bool tgrafo::canonize() {
  for (int i=0; i<ff.size(); i++)
    ff[i].set_colore(0);
  if (canonize_aux()) return TRUE;

  std::list<tgrafo> ls = parenti_minimali(*this, 1);
  for (std::list<tgrafo>::iterator o = ls.begin(); o!=ls.end(); o++)
    if (o->canonize_aux()) {
      *this = *o;
      return TRUE;
    }

  ls = ::disc_replacement(*this,1);
  ls = ::disc_replacement(ls,4);
  for (std::list<tgrafo>::iterator o = ls.begin(); o!=ls.end(); o++) {
    o->newton(1000);
    if (o->sub_canonica()) {
      *this = *o;
      return TRUE;
    }
  }
  for (std::list<tgrafo>::iterator o = ls.begin(); o!=ls.end(); o++)
    if (o->canonize_aux()) {
      *this = *o;
      return TRUE;
    }
  return FALSE;
}

//cerca sub_canonica, se la trova trasforma l'ografo in questa
//funziona bene se il tgrafo non ha facce con colore -1
bool tgrafo::canonize_aux() {
  newton();
  if (sub_canonica()) return TRUE;
  if (!soluzione_geometrica()) return FALSE;
  std::vector<double> tl = tilt();
  for (int lati = 3; lati<=4; lati++) 
    for (int i=0; i<ff.size(); i++) {
      const faccia& f = ff[i];
      if (f.lati() == lati && f.embedded() && f.get_colore()!=-1) {
	bool contiene_tilt_positivo = FALSE;
	for (int t=0; t<lati; t++) 
	  if (tl[v2spigolo(f.get_passaggio(t), f.get_uscita(t))] > tolleranza_tilt)
	    contiene_tilt_positivo = TRUE;
	if (contiene_tilt_positivo)
	  for (int tipo=0; tipo<=lati-3; tipo++) {
	    tgrafo o = *this;
	    int nuova_faccia = o.disc_replacement(i, (bool) tipo);
	    o.ff[nuova_faccia-1].set_colore(-1);
	    if (o.canonize_aux()) {
	      *this = o;
	      return TRUE;
	    }
	  }
      }
    }
  const int l=lung();
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++) {
      int spigolo = v2spigolo(i,j);
      if (tl[spigolo] > tolleranza_tilt) {
	tgrafo o = *this;
	o.MP(i,j);
	if (o.canonize_aux()) {
	  *this = o;
	  return TRUE;
	}
      }
    }
  return FALSE;
}

invarianti::invarianti(const tgrafo& o) {
  vol = o.volume();
  bordo = ::bordo(o);
  angoli = std::vector<std::vector<long double> >(0,std::vector<long double>(0,0));
  for (int i=0; i<o.lung(); i++) {
    std::vector<long double> ang_vertice(6,0);
    for (int j=0; j<6; j++)
      ang_vertice[j] = o.get_theta(i,j);
    angoli.push_back(ang_vertice);
  }
  quante = 1;
}

//controlla che angoli uguali diano volumi uguali
void controllo_volumi(const std::list<tgrafo>& ls) {
  std::list<invarianti> ls_inv = get_invarianti(ls);
  bool con_angoli_prec = con_angoli;
  con_angoli = TRUE;
  for (std::list<invarianti>::const_iterator inv = ls_inv.begin(); inv!=ls_inv.end(); inv++) {
    inv->scrivi();
    for (std::list<invarianti>::const_iterator inv2 = inv; inv2!=ls_inv.end(); inv2++) 
      if (!(*inv2 == *inv) && stessi_angoli(*inv, *inv2)) {
	std::cout << "problema!" << std::endl;
	inv->scrivi(TRUE);
	inv2->scrivi(TRUE);
      }
  }
  con_angoli = con_angoli_prec;
}

std::list<invarianti> get_invarianti(const std::list<tgrafo>& ls) {
  std::list<invarianti> cont;
  for (std::list<tgrafo>::const_iterator o = ls.begin(); o!=ls.end(); o++) {
    if (o->soluzione_geometrica()) {
      invarianti inv(*o);
      inv.ordina();
      bool trovato = FALSE;
      for (std::list<invarianti>::iterator it = cont.begin(); it!=cont.end(); it++)
	if (*it == inv) {
	  trovato = TRUE;
	  it->quante++;
	}
      if (!trovato)
	cont.push_back(inv);
    }
  }
  return cont;
}

bool operator==(const invarianti& inv1, const invarianti& inv2) {
  if (abs(inv1.vol - inv2.vol) > tolleranza_volume || !(inv1.bordo == inv2.bordo))
    return FALSE;
  return (con_angoli ? stessi_angoli(inv1, inv2) : TRUE);
}

bool stessi_angoli(const invarianti& inv1, const invarianti& inv2) {
  if (inv1.angoli.size() != inv2.angoli.size()) return FALSE;
  for (int i=0; i<inv1.angoli.size(); i++)
    for (int j=0; j<6; j++)
      if (abs(inv1.angoli[i][j] - inv2.angoli[i][j]) > tolleranza_theta)
	return FALSE;
  return TRUE;
}

void newton(std::list<tgrafo>& ls) {
  int giro=0;
  for (std::list<tgrafo>::iterator o = ls.begin(); o!=ls.end(); o++) {
    //std::cout << ++giro << std::endl;
    o->newton();
  }
}

void invarianti::ordina() {
  for (int i=0; i<angoli.size(); i++)
    sort(angoli[i].begin(), angoli[i].end());
  sort(angoli.begin(), angoli.end());
}

bool tgrafo::operator<(const tgrafo& o2) const {
  invarianti inv(*this);
  invarianti inv2(o2);
  return (inv < inv2);
}

bool invarianti::operator<(const invarianti& c2) const {
  int cc1 = bordo.componenti_connesse();
  int cc2 = c2.bordo.componenti_connesse();
  if (cc1 < cc2) return TRUE;
  if (cc1 > cc2) return FALSE;	
  int e1 = bordo.get_eulero();
  int e2 = c2.bordo.get_eulero();
  if (e1 < e2) return TRUE;
  if (e1 > e2) return FALSE;
  if (angoli.size() < c2.angoli.size()) return TRUE;
  if (angoli.size() > c2.angoli.size()) return FALSE;
  if (vol < c2.vol) return TRUE;
  if (vol > c2.vol) return FALSE;
  if (con_angoli) for (int i = 0; i<angoli.size(); i++) 
    for (int j=0; j<6; j++) {
      if (angoli[i][j] - c2.angoli[i][j] < -tolleranza_theta)
	return TRUE;
      if (angoli[i][j] - c2.angoli[i][j] > tolleranza_theta)
	return FALSE;
    }
  return FALSE;
}

void invarianti::scrivi(bool scrivi_angoli) const {
  std::cout << "Quante = " << quante;
  std::cout << "   tetraedri = " << angoli.size();
  std::cout << "    Vol = " << vol << "    bordo = ";
  bordo.scrivi(std::cout);
  if (scrivi_angoli) {
    std::cout << "Angoli: " << std::endl;
    for (int i=0; i<angoli.size(); i++) {
      for (int j=0; j<6; j++)
	std::cout << angoli[i][j] << "   ";
      std::cout << std::endl;
    }
  }
}
