#if !defined (GRAPH_H)
#define GRAPH_H

#include<vector>
#include<list>
#include "defs.h"

class torsione{
  int t;
 public:
  int get_t() const {return t;}
  void set_t(int ts) {t=ts==3 ? ts:ts % 3;}
  bool falsa() const {return (t==3);}
  torsione(int ts=-1):t(ts==3 ? ts:ts % 3) {}
  inline torsione& operator+=(int s) {if (t!=3) t = (t+s) % 3; return *this;}
	inline torsione& operator++() {if (t!=3) t = (t+1) % 3; return *this;} 
  void invert() {if (t!=3) t=(3-t)%3;}
	torsione operator+=(const torsione& t);
  friend bool operator==(const torsione& x, const torsione& y);
  friend bool operator!=(const torsione& x, const torsione& y);
  friend torsione operator+(const torsione& x, const torsione& y);
  void scrivi(std::ostream& f) const;
};

//vertice quadrivalente. E' visto a meno di permutazioni della quadrupla
//vicino. Le permutazioni diventano pero' signifative nelle classi derivate.
class vertice {
 protected:
  int vicino[4];
 public:
  vertice();
  vertice(const int*);
  int get_vicino(int i) const {return vicino[i];}
  void set_vicino(int i, int j) {vicino[i]=j;}
  int tipo(int) const;
  bool segno() const;

  virtual vertice* clone() const {return new vertice(*this);}
	virtual ~vertice() {}
  virtual bool operator==(const vertice& y) const;
  virtual bool operator<(const vertice& y) const; 

	//hanno senso nella classe ivertice:
  virtual int get_col(int i) const {return -1;}
  virtual void set_col(int i, int cl) {}

  //hanno senso nella classe overtice:
  virtual void set_tor(int i, torsione tr) {}
  virtual torsione get_tor(int i) const {return 0;}
  virtual void tor_add(int i, int tr) {}
  virtual void tor_invert(int i) {}
  virtual bool falso(int i) const {return 0;}
};

class trasformazione {
 public:
  tensor2 t;
  trasformazione():t() {};
};

class grafo {
 protected:
  std::vector<vertice*> v;
 public:
  grafo():v() {}
  grafo(const grafo& g);
  grafo(const tensor2& w, int l);
  grafo(std::istream&, int tipo_formato = 0);
  grafo(int, long);
  virtual ~grafo();
  
  grafo& operator=(const grafo& g);
  friend bool operator==(const grafo& x, const grafo& y);
  const vertice* operator[] (int i) const {return v[i];}
  virtual grafo* clone() const {return new grafo(*this);}
  virtual bool operator<(const grafo& y) const;
  
  int get_v(int i, int k) const {return v[i]->get_vicino(k);}
  int lung() const {return v.size();}
  int quanti_spigoli(int vert1, int vert2) const;
  
  virtual void trasforma(const trasformazione&);
  virtual trasformazione trasformazione_banale() const;
  virtual std::list<trasformazione> trasformazioni_minimizzanti(bool fermati_subito = FALSE) const;
  
  bool connesso(const tensor2&) const; 
  int tipo(int) const;
  
  bool set_forma_minima();
  bool forma_minima() const;

  bool utile() const;
  bool irriducibile() const;
  bool riccio(int i) const;
  
  virtual void scrivi(std::ostream&, int tipo_formato = 0) const;
  virtual void salva(std::ostream&, int tipo_formato = 0) const;
  
 protected:
  void aggiungi_vertici(int);
  void trasformazioni_minimizzanti(std::list<trasformazione>& trasfmin, int liv, bool fermati_subito) const;  

 private:
  void utile(tensor2&, int[][4], int[][4], int, int, bool&) const;
};

void set_grafi (int, bool solo_utili = FALSE);
void set_grafi (int, int, int, tensor2&, tensor2&, std::ostream&, bool);
void set_grafi_utili(int, int, int, tensor2&, std::vector<int>&, std::ostream&);
void set_utili(int);

tensor3 cicli(const grafo& g, int max);
tensor2 catena(const grafo& g, int vert, int max=0);

bool operator==(const trasformazione& t1, const trasformazione& t2);

#endif
