#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <list>
#include <algorithm>
#include <iomanip>

#include "defs.h"
#include "math.h"
#include "templates.cpp"
#include "graph.h"
#include "igraph.h"

extern const int Aquattro[12][4] = {{0,1,2,3},{1,0,3,2},{2,3,0,1},{3,2,1,0},{0,2,3,1},{0,3,1,2},{2,1,3,0},{3,1,0,2},{1,3,2,0},{3,0,2,1},{1,2,0,3},{2,0,1,3}};
extern const int Aquattroinv[12][4] = {{0,1,2,3},{1,0,3,2},{2,3,0,1},{3,2,1,0},{0,3,1,2},{0,2,3,1},{3,1,0,2},{2,1,3,0},{3,0,2,1},{1,3,2,0},{2,0,1,3},{1,2,0,3}};

extern const int RR[4] = {1,2,3,0};

//costruttore standard. Mette col = {-1,-1,-1,-1}
ivertice::ivertice():vertice() {
  for (int i=0; i<4; i++)
    col[i] = -1;
}

//costruisce da vertice, con collocazioni date da cl
ivertice::ivertice(const vertice& v, const int* cl):vertice(v) {
  for (int i=0; i<4; i++)
    col[i] = cl[i];
}

//costruttore di copia
igrafo::igrafo (const igrafo& g) {
	for (int i=0; i<g.lung(); i++)
		v.push_back(g[i]->clone());
}

//costruisce a partire di g: da ogni vertice di g costruisce un vertice con vicino[] ordinato.
//Quindi fa i collegamenti nel modo naturale, cioe' se ci sono piu' collegamenti fra i due vertici
//mette il primo col primo, il secondo col secondo, etc.
//Infine, se or[vert]==1 allora scambia le prime due col[] del vertice vert.
//
//Importante: se un igrafo e' in forma minima, allora si ottiene come igrafo(g, or)
//per un 'or' opportuno.
igrafo::igrafo(const grafo& g, const std::vector<int>& ori) {
  int l = g.lung();
  std::vector<int> quanti_gia_collegati(l,0);
  int temp[4] = {-1,-1,-1,-1};
  for (int i=0; i<l; i++)
    v.push_back(new ivertice());
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      if (get_col(i,j)==-1) {
				int i2 = g.get_v(i,j);
				if (i2!=i) set_col(i,j,i2,quanti_gia_collegati[g.get_v(i,j)]++);
				else set_col(i,j,i2,j+1);
			}
  const int z[4] = {0,1,3,2};
  for (int i=0; i<l; i++)
    if (ori[i])
			Cz(i, z);
}

//collega i vertici i e j con uscite i2 e j2
void igrafo::set_col(int i, int j, int i2, int j2, torsione tr) {
  v[i]->set_vicino(j,i2);
  v[i2]->set_vicino(j2,i);
  v[i]->set_col(j,j2);
  v[i2]->set_col(j2,j);
	set_tor(i,j,tr);
}

//assegna la torsione tr al vertice (i,j)
void igrafo::set_tor(int i, int j, torsione tr) {
  v[i]->set_tor(j,tr); v[get_v(i,j)]->set_tor(get_col(i,j),tr);
}

//aggiunge 'k' alla torsione del vertice (i,j)
void igrafo::tor_add(int i, int j, int k) {
  v[i]->tor_add(j,k); v[get_v(i,j)]->tor_add(get_col(i,j),k);
}

//moltiplica la torsione del vertice (i,j) per -1
void igrafo::tor_invert(int i, int j) {
  v[i]->tor_invert(j); v[get_v(i,j)]->tor_invert(get_col(i,j));
}

//permuta le uscite del vertice 'vert' secondo la permutazione 'z'
//cioe': lo spigolo collegato all'uscita i e' spostato all'uscita z[i].
void igrafo::Cz(int vert, const int* z, int delta_tor) {
  vertice* temp = v[vert]->clone();
  for (int i=0; i<4; i++) {
		int vert2 = temp->get_vicino(i);
		int col2 = temp->get_col(i);
		if (vert2 == vert) col2 = z[col2];
    set_col(vert, z[i], vert2, col2, temp->get_tor(i));
	}
  for (int i=0; i<4; i++)
    tor_add(vert,i,delta_tor);
  delete temp;
}

//mossa R sul vertice 'vert'
void igrafo::R(int vert) {
  Cz(vert, RR);
}

//mossa C sul vertice 'vert', di tipo 'tipo', invertita se 'inverso'==TRUE
void igrafo::C(int vert, int tipo, int inverso) {
  int z[4]; 
  for (int t=0; t<4; t++) inverso ? z[t] = Aquattroinv[tipo][t] : z[t] = Aquattro[tipo][t];
  Cz(vert, z, effetto_C_su_tor(tipo, inverso));
}

//fa le mosse C sui vertici, come scritto su 'mosse'
void igrafo::C(const std::vector<int>& mosse) {
	int l = lung();
	for (int t=0; t<l; t++)
		C(t,mosse[t]);
}

// si riflette rispetto al piano su cui giace
void igrafo::specchio () {
  const int l = lung();
  for (int i=0; i<l; i++)
    R(i);
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      if (prima_occorrenza_spigolo(i,j))
        tor_invert(i,j);
}

//minimizza con mosse C
std::vector<int> igrafo::minimizza_con_C() {
  std::vector<int> mosse = mosse_C_minimizzanti();
	C(mosse);
	return mosse;
}

//Restituisce, tra tutti i vettori di mosse C, quello (o uno di quelli) che
//produce l'igrafo minore.
std::vector<int> igrafo::mosse_C_minimizzanti() const {
  std::vector<int> mosse_min(lung(),0);
  mosse_C_minimizzanti(0, mosse_min);
  return mosse_min;
}

//Ricorsiva, serve a minimizza_con_C()
//(questa e' private per impedire che venga chiamata con q>0).
void igrafo::mosse_C_minimizzanti (int q, std::vector<int>& mosse_min) const {
  static igrafo* modello;
  static int l;
  static std::vector<int> mosse;
  if (q==0) {
    modello = clone();
    l = lung();
    mosse_min = std::vector<int>(l,0);
    mosse = std::vector<int>(l,0);
  }
  for (int i=0; i<12; i++) {
    modello->C(q,i,0); 
    int z[4]; for (int t=0; t<4; t++) z[t]=modello->get_v(q,t);
    bool ok = TRUE;
    
    //controlla che la configurazione non sia banalmente non minimale
    int tp = tipo(q);
    if (tp==-1 && !(z[0]<=z[1] && z[1]<=z[2] && z[1]<=z[3])) ok = FALSE;
    if (tp!=-1 && !(z[0]<=z[1] && z[1]<=z[2] && z[2]<=z[3])) ok = FALSE;
    if (tp==2 && z[1]<q && ok) {
      int diverso = 3*(modello->get_v(z[1],3)!=q); //get_v(z[1]) e' xqqq (diverso=0) o qqqx (diverso=3)
      int col[3]; for (int t=0; t<3; t++) col[t] = modello->get_col(z[1],(diverso+1+t) % 4);
      if (col[0]>col[1] || col[0]>col[2]) ok = FALSE;
    }    
    if (ok) {  
      mosse[q]=i;
      if (q==l-1) {
				igrafo* modello_min = clone();
				modello_min->C(mosse_min);
        if (*modello < *modello_min)
          mosse_min = mosse;
        delete modello_min;
      }
      else mosse_C_minimizzanti(q+1, mosse_min);
    }
    modello->C(q,i,1);
  }
  if (q==0)
    delete modello;
}

//se fermati_subito==FALSE (default): restituisce tutte le trasformazioni che trasformano l'igrafo
//  in forma minima.
//ATTENZIONE: non le restituisce tutte, ma al piu' una per ogni permutazione e specchio!!
//se fermati_subito==TRUE: restituisce una trasformazione che trasforma il il grafo in un grafo
//  minore, se c'e'. Altrimenti restituisce la trasformazione banale.
//in ogni caso, restituisce una lista non vuota, e se contiene la trasformazione banale
//allora e' in testa
std::list<trasformazione> igrafo::trasformazioni_minimizzanti(bool fermati_subito) const {
  std::list<trasformazione> perm_lista = grafo::trasformazioni_minimizzanti(FALSE);
  std::list<trasformazione> trasfmin = std::list<trasformazione> (1,trasformazione_banale());
  const int l = lung();
  for (std::list<trasformazione>::iterator perm = perm_lista.begin(); perm!=perm_lista.end(); perm++) {
    for (perm->t[1][0]=0; perm->t[1][0]<=1; perm->t[1][0]++) {
      igrafo* nuovo = clone();
      nuovo->trasforma(*perm);
      perm->t[2] = nuovo->minimizza_con_C();
      igrafo* min = clone();
      min->trasforma(trasfmin.front());
      if (*nuovo < *min) {
        trasfmin = std::list<trasformazione>(1,*perm);
        if (fermati_subito) return trasfmin;
      }
      if (*nuovo == *min && !std::count(trasfmin.begin(), trasfmin.end(), *perm))
        trasfmin.push_back(*perm);
      delete nuovo; delete min;
      perm->t[2] = std::vector<int>(l,0);
    }
  }
  return trasfmin;
}

//restituisce la trasformazione banale (cioe' l'identita')
trasformazione igrafo::trasformazione_banale() const {
  const int l = lung();
  trasformazione trasf;
  trasf.t.push_back(std::vector<int>(l,0));
  for (int i=0; i<l; i++) trasf.t[0][i] = i;
  trasf.t.push_back(std::vector<int>(1,0));
  trasf.t.push_back(std::vector<int>(l,0));
  return trasf;
}

//trasforma l'igrafo secondo 'trasf'.
void igrafo::trasforma(const trasformazione& trasf) {
  grafo::trasforma(trasf);
  if (trasf.t[1][0]) specchio();
	C(trasf.t[2]);
}

tensor2 mosse_C (const igrafo x, const igrafo y, int q);

//restituisce tutti i vettori di mosse C che trasformano x in y.
//Ovvero, per ciascun vettore v, operando sul vertice i di x una mossa C(v[i])
//per ogni i, si ottiene y.
//Non so se puo' esistere piu' di un vettore!
//DOMANDA: servono tutti o ne basta uno ?
tensor2 mosse_C (const igrafo& x, const igrafo& y) {
  return mosse_C(x,y,0);
}

//ricorsiva, serve a mosse_C. E' private.
tensor2 mosse_C (const igrafo x, const igrafo y, int q) {
  static igrafo modello;
  static int l;
  static std::vector<int> mosse;
  static tensor2 ls;
  if (q==0) {
    modello = x;
    l = x.lung();
    mosse = std::vector<int>(l);
    ls = tensor2();
  }
  else for (int i=0; i<12; i++) {
    modello.C(q,i,0);
    bool ammissibile = TRUE;
    for (int j=0; j<4; j++)
      if (modello.get_v(i,j)!=y.get_v(i,j))
        ammissibile = FALSE;
    if (ammissibile) {
      mosse[q]=i;
      if (q==l-1 && modello == y)
        ls.push_back(mosse);
      else mosse_C(x,y,q+1);
    }
    modello.C(q,i,1);
  }
}

// dice se (vert,usc)<(vert',usc') lessicografico, ove vert',usc' determina lo stesso spigolo.
bool igrafo::prima_occorrenza_spigolo(int vert, int usc) const {
  return (get_v(vert, usc) > vert || 
          (get_v(vert, usc)==vert && get_col(vert,usc)>usc));
}
  
// (vert,usc)-->spigolo	 (che parte da zero)
int igrafo::v2spigolo(int vert, int usc) const {
  int vert2 = get_v(vert, usc);
  int usc2 = get_col(vert, usc);
  if (vert2<vert || vert2==vert && usc2<usc) {vert=vert2; usc=usc2;}
  int spigolo=0;
  for (int i=0; i<v.size(); i++)
    for (int j=0; j<4; j++)
      if (get_v(i,j)>i || get_v(i,j)==i && get_col(i,j)>j) {
        if (i==vert && j==usc) return spigolo;
        spigolo++;
      }
  return -1;
}

// spigolo-->(vert,usc)
void igrafo::spigolo2v (int s, int& vert, int& col) const {
  int k=0;
  for (int i=0; i<v.size(); i++)
    for (int j=0; j<4; j++)
      if (get_v(i,j)>i || get_v(i,j) == i && get_col(i,j)>j) {
        if (k==s) {vert=i; col=j; return;}
        k++;
      }
}

void igrafo::scrivi(std::ostream& f, int tipo_formato) const{
  const int l=lung();
  f << "N.Vertici: ";
  f << l << "\n";
  for (int i=0; i<l ; i++) {
    f << "N. " << i;
    if (i<10) f << " ";
    f << ":   ";
    for (int j=0; j<4; j++)
      f << std::setw(2) << get_v(i,j);
    f << "    ";
    for (int j=0; j<4; j++)
      f << std::setw(2) << get_col(i,j);
    f << "\n";
  }
}

std::list<igrafo> crea_igrafi(const grafo& g, int liv);

//restituisce la lista di igrafi ottenibili da g, a meno di automorfismi
std::list<igrafo> crea_igrafi(const grafo& g) {
  return crea_igrafi(g, 0);
}

//funzione ricorsiva che serve per 'crea_igrafi'.
std::list<igrafo> crea_igrafi(const grafo& g, int liv) {
  static int l;
  static std::list<igrafo> ls;
  static std::vector<int> ori;
  static std::vector<int> solo_zero;
  if (liv==0) {
    l = g.lung();
    ls = std::list<igrafo>();
    ori = std::vector<int>(l,0);
		solo_zero = std::vector<int>(1,0); //grazie a mossa R abbiamo ori[0]=0
		for (int i = 0; i<l; i++)
			if (g.riccio(i)) solo_zero.push_back(i);
    for (int i = 0; i<l; i++)
			for (int i2 = i+1; i2<l; i2++) {
				int qs = g.quanti_spigoli(i,i2);
				if (qs>=2) solo_zero.push_back(i);
			}
  }
  if (liv<l) {                     
    bool only_zero = std::count(solo_zero.begin(), solo_zero.end(), liv)>0;
    for (int orientaz=0; orientaz <= 1-only_zero; orientaz++) {
      ori[liv] = orientaz;
      crea_igrafi(g, liv+1);
    }
  }
  else {
    igrafo ig = igrafo(g, ori);

		//non tutti gli igrafi in forma minima vengono presi cosi'! cio' dipende infatti
		//dal vettore z[4] dentro igrafo(g,ori)
		ig.set_forma_minima();
		if (std::count(ls.begin(), ls.end(), ig) == 0) 
			ls.push_back(ig);
  }
	//elimina_poliedri_doppi(ls);
	
  return ls;
}

// x<y lessicografico su (grafo, vicini, cols);
bool igrafo::operator<(const grafo& yg) const {
  const igrafo& y = (igrafo&) yg;
  const int l = lung();
  if ((grafo) *this < (grafo) y) return TRUE;
  if ((grafo) y < (grafo) *this) return FALSE;
	for (int i=0; i<l; i++)
    for (int j=0; j<4; j++) {
      int xvic = get_v(i,j);
      int yvic = y.get_v(i,j);
      if (xvic < yvic) return TRUE;
      if (xvic > yvic) return FALSE;
    }
	for (int i=0; i<l; i++)
    for (int j=0; j<4; j++) {
      int xcol = get_col(i,j);
      int ycol = y.get_col(i,j);
      if (xcol < ycol) return TRUE;
      if (xcol > ycol) return FALSE;
    }
  return FALSE;
}

// x==y per ivertici
bool ivertice::operator==(const vertice& y) const {
  for (int t=0; t<4; t++) {
    if (get_vicino(t)!=y.get_vicino(t)) return FALSE;
    if (get_col(t)!=y.get_col(t)) return FALSE;
  }
  return TRUE;
}

//assegnamento
igrafo& igrafo::operator=(const igrafo& g) {
	disalloca(v);
	v.clear();
	for (int i=0; i<g.lung(); i++)
		v.push_back(g[i]->clone());
	return *this;
}

//restituisce un intero tr tra 0 e 2. L'effetto di una mossa C(tipo, inverso) 
//su uno spigolo adiacente una
//volta al vertice e', sulla torsione, tor_add(tr).
int effetto_C_su_tor(int tipo, int inverso) {
  int z[4]; for (int t=0; t<4; t++) inverso ? z[t] = Aquattroinv[tipo][t] : z[t] = Aquattro[tipo][t];
  int fis=0, tr=0, pt=0;
  for (int i=0; i<4; i++) if (z[i]==i) {pt=i; fis++;}
  if (fis==1) {
    if (z[mod(pt+1,4)] == mod(pt+3,4)) tr=1; else tr=2;
    if (mod(pt,2)) tr=3-tr;
  }
  return tr;
}
