#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <vector>

#include "io.h"
#include "surfaces.h"
#include "groups.h"

extern bool mostra_contatore;
bool inglese = TRUE;

void scrivi(const grafo& o, std::ostream& f, int tipo_formato) {
	o.scrivi(f, tipo_formato);
}

//scrive su f i dati di o (o di una sua classe derivata)
//e i dati topologici di o (cosa che o.scrivi(f) non fa)
void scrivi(const ografo& o, std::ostream& f, int tipo_formato) {
	o.scrivi(f, tipo_formato);
	o.scrivi_facce(f); 
	o.scrivi_geometria(f);
	if (o.spigoli_falsi().empty()) {
		f << (inglese ? "Boundary: " : "Bordo: "); 
		bordo(o).scrivi(f);
		f << "Pi: ";
		gruppo p=pi(o).semplifica().forma_standard();
		p.scrivi(f);
		f << (inglese ? "Homology: " : "Omologia: ");
		abelianizzato(p).scrivi(f);
	}
	/* ATTENZIONE
	f << (inglese ? "Pseudominimality: " : "Pseudominimale: ");
	string ris;
	o.pseudo_minimale(&ris);
	f << ris;
	*/

	if (o.scheletro()) {
		f << (inglese ? "Faces incident to 'fake' edges: " : 
					"Facce i incidenti su spigoli falsi: ");   //parte da 1
		int num=1;
		int ns;
		std::vector<int> spigoli = o.spigoli_falsi();
		for (int num = 0; num < spigoli.size(); num++) {
			if (num>0) f << ", ";
			int i,j; o.spigolo2v(spigoli[num],i,j);
			teta th = abs(o.facce_su_spigolo(i,j)); th.scrivi(f);
		}
		f << (inglese ? "\nTheta-graphs: " : "\nTriodi: ");
		tensor2 trd = o.triodi();
		for (int t=0; t<trd.size(); t++) {
			f << "{";
			for (int t2=0; t2<3; t2++) {
				f << trd[t][t2]; if (t2<2) f << ",";
			}
			f << "} ";
		}
		f << (inglese ? "\nEdges containing Moebius triplets: " : 
					"\nSpigoli contenenti Mobius triplet: ");
		spigoli = o.spigoli_con_mobius_triplet();
		for (int t=0; t<spigoli.size(); t++) f << spigoli[t] << " ";
		f << "\n";
	}
	/* ATTENZIONE
	int num_1scheletro = o.num_1scheletro();
	if (num_1scheletro) f << (inglese ? "1-skeleton number: " : 
														"Numero 1-scheletro: ") << num_1scheletro << "\n";
	*/
	f << "-------------------\n";
}

void scrivi(const std::vector<int>& v) {
	for (int i=0; i<v.size(); i++)
		std::cout << v[i] << "  ";
	std::cout << "\n";
}

void scrivi(const tensor2& v) {
	for (int i=0; i<v.size(); i++)
		scrivi(v[i]);
}

void scrivi(const tensor3& v) {
	for (int i=0; i<v.size(); i++) {
		scrivi(v[i]);
		std::cout << "\n";
	}
}

